#ifndef __TXP__BACKENDS__PORTABLEDOCUMENTFORMAT_H
    #define __TXP__BACKENDS__PORTABLEDOCUMENTFORMAT_H


    // developed with the help of
    //  -  "PDF Explained" by John Whitington (O’Reilly)
    //     Copyright 2012 John Whitington, ISBN: 978-1-449-31002-8
    //  -  http://brendanzagaeski.appspot.com/0004.html
    //  -  http://www.adobe.com/devnet/pdf/pdf_reference.html



    #include<list>
    #include<sstream>
    #include<iomanip>
    #include<map>
    #include<set>
    #include<vector>
    #include "backends.h"
    #include "postscript.h"

    class PortableDocumentFormatPageLink
    {
        public:
            int x;
            int y;
            int w;
            int h;
            int page;
            PortableDocumentFormatPageLink(int x, int y, int w, int h, int page);
    };

    class PortableDocumentFormatHyperlink
    {
        public:
            int x;
            int y;
            int w;
            int h;
            std::string url;
            PortableDocumentFormatHyperlink(int x, int y, int w, int h, const std::string& url);
    };

    class PortableDocumentFormatBookmark
    {
        public:
            int ObjectId;

            PortableDocumentFormatBookmark* Parent;
            PortableDocumentFormatBookmark* FirstChild;
            PortableDocumentFormatBookmark* LastChild;
            PortableDocumentFormatBookmark* PrevSibling;
            PortableDocumentFormatBookmark* NextSibling;

            std::string Caption;
            int Page;

            PortableDocumentFormatBookmark(int Page, const std::string& Caption, PortableDocumentFormatBookmark* Parent = NULL);
    };

    class TXPBackendPortableDocumentFormat : public TXPBackendFile
    {
        protected:
            std::map<std::string, int> FontIds;
            std::list<std::string> Fonts;
            std::map<std::string, std::vector<wchar_t> > Codepage;
            std::map<std::string, std::map<wchar_t,int> > CodepageLookup;

            std::list<PortableDocumentFormatPageLink> PageLinks;
            std::list<PortableDocumentFormatHyperlink> Hyperlinks;
            std::vector<PortableDocumentFormatBookmark*> Bookmarks;

            std::list<int> ObjectOffsets;
            int ObjectCounter;
            int FontRegisterObjectId;
            int PageTreeObjectId;
            int InfoObjectId;
            int CatalogObjectId;
            int FirstPageObjectId;
            std::stringstream Page;

            std::string ToPdfSystemString(const std::wstring& str);
            std::string ToPdfString(const std::string& fontname, const std::wstring& str);
        public:
            TXPBackendPortableDocumentFormat(std::string FileName);
            ~TXPBackendPortableDocumentFormat();

            void Initialize(int NumberOfPages, time_t CreationTime, std::wstring Author=L"", std::wstring Title=L"");
            void NewPage(int PageNumber, float width_cm, float height_cm, float resolution_dpi);
        protected:
            void PutTextDots(int left, int top, const std::wstring& text, TXPFontSettings* font);
            void PutPageLinkDots(int leftDots, int topDots, int widthDots, int heightDots, int PageNumber);
            void PutHyperlinkDots(int leftDots, int topDots, int widthDots, int heightDots, const std::string& url);
            void PutLineDots(int x1Dots, int y1Dots, int x2Dots, int y2Dots);
        public:
            int PutBookmark(int PageNumber, const std::wstring& Caption, int ParentId = -1);
            void FinishPage();
            void Finish();
    };


#endif

