
#include "../../include/backends/portabledocumentformat.h"
using namespace std;


            // !!!!!!!!!!!!!!! change when number of objects per page changes!!!
            const int ObjectsPerPage = 2;


/*
        void GraphRendererPDF::SetPenColor(Visualization::Color color)
        {
            oss << (color.Red/256.0f) << " " << (color.Green/256.0f) << " " << (color.Blue/256.0f) << " RG\n";
        }

        void GraphRendererPDF::SetBrushColor(Visualization::Color color)
        {
            oss << (color.Red/256.0f) << " " << (color.Green/256.0f) << " " << (color.Blue/256.0f) << " rg\n";
        }

        void GraphRendererPDF::SetLineWidth(float width)
        {
            oss << (int)(width*ResolutionDPCM) << " w\n";
        }

        void GraphRendererPDF::Circle(float x, float y, float Radius)
        {
            // http://forums.adobe.com/message/2312083
            // WTF, Adobe!? Seriously, WTF!?

            int radius = Radius * ResolutionDPCM;
            int h = 0.5522422 * Radius * ResolutionDPCM; // sine of 33 deg 31 sec

            oss << "q 1 0 0 1 " << x*ResolutionDPCM << " " << (ImageHeight-y)*ResolutionDPCM << " cm\n"
                << "0 " << -radius << " m\n"

                << h << " " << -radius << " " << radius << " " << -h << " " << radius << " " << 0 << "c\n"
                << radius << " " << h << " " << h << " " << radius << " " << 0 << " " << radius << "c\n"
                << -h << " " << radius << " " << -radius << " " << h << " " << -radius << " " << 0 << "c\n"
                << -radius << " " << -h << " " << -h << " " << -radius << " " << 0 << " " << -radius << "c\n"
                << "f Q\n";
        }

 */


    std::string TXPBackendPortableDocumentFormat::ToPdfSystemString(const std::wstring& str)
    {
        std::string result = "(";
        bool unicode = false;
        for(std::wstring::const_iterator i = str.begin(); i != str.end() && !unicode; i++)
        {
            if((unsigned int)(*i) > 127)
                unicode = true;
            switch(*i)
            {
                case '\\': result += "\\\\";     break;
                case '(':  result += "\\(";      break;
                case ')':  result += "\\)";      break;
                default:   result += (char)(*i); break;
            }
        }
        result += ")";

        if(unicode)
        {
            static std::string hex = "0123456789ABCDEF";
            result = "<FEFF";

            for(std::wstring::const_iterator i = str.begin(); i != str.end(); i++)
            {
                unsigned int id = (unsigned int)(*i);
                if(id > 0xFFFF) // pdf can handle only 16-bit encoded unicode (pathetic)
                    continue;
                result += hex[(id & 0xF000) >> 12];
                result += hex[(id & 0x0F00) >> 8];
                result += hex[(id & 0x00F0) >> 4];
                result += hex[(id & 0x000F)];
            }
            result += ">";
        }
        return result;
    }



    std::string TXPBackendPortableDocumentFormat::ToPdfString(const std::string& fontname, const std::wstring& str)
    {
        std::string result = "<";
        static std::string hex = "0123456789ABCDEF";
        for(std::wstring::const_iterator i = str.begin(); i != str.end(); i++)
        {
            int id;
            if(CodepageLookup[fontname].find(*i) == CodepageLookup[fontname].end())
            {
                id = Codepage[fontname].size();
                Codepage[fontname].push_back(*i);
                CodepageLookup[fontname][*i] = id;
            }
            else
            {
                id = CodepageLookup[fontname][*i];
            }
            ++id;
            result += hex[(id & 0x00F0) >> 4];
            result += hex[(id & 0x000F)];
        }
        result += ">";
        return result;
    }


    PortableDocumentFormatPageLink::PortableDocumentFormatPageLink(int x, int y, int w, int h, int page)
    {
        this->x = x;
        this->y = y;
        this->w = w;
        this->h = h;
        this->page = page-1;
    }

    PortableDocumentFormatHyperlink::PortableDocumentFormatHyperlink(int x, int y, int w, int h, const std::string& url)
    {
        this->x = x;
        this->y = y;
        this->w = w;
        this->h = h;
        this->url = url;
    }

    PortableDocumentFormatBookmark::PortableDocumentFormatBookmark(int Page, const std::string& Caption, PortableDocumentFormatBookmark* Parent)
    {
        this->Page = Page-1;
        this->Caption = Caption;
        this->Parent = Parent;

        FirstChild = NULL;
        LastChild = NULL;
        PrevSibling = NULL;
        NextSibling = NULL;

        if(Parent != NULL)
        {
            if(Parent->FirstChild == NULL)
            {
                Parent->FirstChild = this;
                Parent->LastChild = this;
            }
            else
            {
                PrevSibling = Parent->LastChild;
                PrevSibling->NextSibling = this;
                Parent->LastChild = this;
            }
        }
    }



    TXPBackendPortableDocumentFormat::TXPBackendPortableDocumentFormat(std::string FileName)
        : TXPBackendFile(FileName)
    {
        BottomUpYAxis = true;
        CurrentPageResolution_dpi = 72.0f;
        ObjectCounter = 1;
        PageTreeObjectId = -1;
        InfoObjectId = -1;
        CatalogObjectId = -1;
        FontRegisterObjectId = -1;

        Bookmarks.push_back(new PortableDocumentFormatBookmark(-1, "", NULL));
    }

    TXPBackendPortableDocumentFormat::~TXPBackendPortableDocumentFormat()
    {
        for(std::vector<PortableDocumentFormatBookmark*>::iterator i = Bookmarks.begin(); i != Bookmarks.end(); i++)
            delete *i;
    }

    void TXPBackendPortableDocumentFormat::Initialize(int NumberOfPages,
        time_t CreationTime, std::wstring Author, std::wstring Title)
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::Initialize\n";

        TXPBackendFile::Initialize(NumberOfPages, CreationTime, Author, Title);

        ObjectCounter = 1;
        (*stream) << "%PDF-1.7\n";
        (*stream) << "%\xE2\xE3\xCF\xD3\n";

        // Object 1 for Meta-Infos of the file
        ObjectOffsets.push_back(stream->tellp());
        (*stream) << ObjectCounter << " 0 obj \n";
        InfoObjectId = ObjectCounter;
        (*stream) << "<<\n";
        if(Title.size() > 0)
            (*stream) << "  /Title " << ToPdfSystemString(Title) << "\n";
        if(Author.size() > 0)
            (*stream) << "  /Author " << ToPdfSystemString(Author) << "\n";
        struct tm* t = localtime(&CreationTime);
        (*stream) << "  /CreationDate (D:" << setw(4) << setfill('0') << (t->tm_year+1900)
                                     << setw(2) << setfill('0') << (t->tm_mon+1)
                                     << setw(2) << setfill('0') << (t->tm_mday)
                                     << setw(2) << setfill('0') << (t->tm_hour)
                                     << setw(2) << setfill('0') << (t->tm_min)
                                     << setw(2) << setfill('0') << (t->tm_sec)
                                     << ")\n";
        /*
        (*stream) << "  /ModDate (D:" << setw(4) << setfill('0') << (t->tm_year+1900)
                                     << setw(2) << setfill('0') << (t->tm_mon+1)
                                     << setw(2) << setfill('0') << (t->tm_mday)
                                     << setw(2) << setfill('0') << (t->tm_hour)
                                     << setw(2) << setfill('0') << (t->tm_min)
                                     << setw(2) << setfill('0') << (t->tm_sec)
                                     << ")\n";
        */
        (*stream) << "  /Creator (LaTeX++ by Viktor \"AlgorithMan\" Engelmann)\n";
        (*stream) << ">>\n";
        (*stream) << "endobj \n\n";
        ObjectCounter++;


        FirstPageObjectId = ObjectCounter+1;
        FontRegisterObjectId = FirstPageObjectId + ObjectsPerPage*NumberOfPages;


        // Object 2 is the Page-Tree
        ObjectOffsets.push_back(stream->tellp());
        PageTreeObjectId = ObjectCounter;
        (*stream) << ObjectCounter << " 0 obj \n";
        (*stream) << "<<\n";
        (*stream) << "  /Type /Pages\n";
        (*stream) << "  /Count " << NumberOfPages << "\n";
        (*stream) << "  /Kids [";
        for(int i = 0; i < NumberOfPages; i++)
            (*stream) << FirstPageObjectId+i*ObjectsPerPage << " 0 R ";
        (*stream) << "]\n";
        (*stream) << ">>\n";
        (*stream) << "endobj \n\n% ==================================================================================== PAGES\n\n";
        ObjectCounter++;

    }


    void TXPBackendPortableDocumentFormat::NewPage(int PageNumber, float width_cm, float height_cm, float)
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::NewPage\n";

        TXPBackendFile::NewPage(PageNumber, width_cm, height_cm, 72.0f);

        // Object 3 is the declaration of Page 1
        ObjectOffsets.push_back(stream->tellp());
        (*stream) << ObjectCounter << " 0 obj \n";
        (*stream) << "<<\n";
        (*stream) << "  /Type /Page\n";
        (*stream) << "  /Parent " << PageTreeObjectId << " 0 R\n";
        (*stream) << "  /MediaBox [0 0 " << (int)(width_cm*CurrentPageResolution_dpi/2.54)
                  << " " << (int)(height_cm*CurrentPageResolution_dpi/2.54) << "]\n";
    }


    void TXPBackendPortableDocumentFormat::PutTextDots(int leftDots, int topDots, const std::wstring& text,
        TXPFontSettings* font)
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::PutTextDots\n";

        if(FontIds.find(font->Font->Name) == FontIds.end()) // font used for the first time
        {
            Fonts.push_back(font->Font->Name);
            FontIds[font->Font->Name] = Fonts.size();
        }

        Page << "BT"
             << " /F" << FontIds[font->Font->Name] << " " << font->FontSize << " Tf"
             << " " << leftDots << " " << topDots << " Td"
             << " " << ToPdfString(font->Font->Name, text) << " Tj"
             << " ET\n";
    }

    void TXPBackendPortableDocumentFormat::PutPageLinkDots(int leftDots, int topDots, int widthDots, int heightDots, int PageLink)
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::PutPageLinkDots\n";

        PageLinks.push_back(PortableDocumentFormatPageLink(leftDots,topDots,widthDots,heightDots,PageLink));
    }

    void TXPBackendPortableDocumentFormat::PutHyperlinkDots(int leftDots, int topDots, int widthDots, int heightDots, const std::string& url)
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::PutHyperlinkDots\n";

        Hyperlinks.push_back(PortableDocumentFormatHyperlink(leftDots,topDots,widthDots,heightDots, url));
    }

    void TXPBackendPortableDocumentFormat::PutLineDots(int x1Dots, int y1Dots, int x2Dots, int y2Dots)
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::PutLineDots\n";

        Page << "BX /GS1 gs EX\n"; // ???? vgl ExtGState

        Page << "q\n";
        Page << "2 w\n";
        Page << "0 0 0 rg\n";
        Page << "0 0 0 RG\n";
        Page << x1Dots << " " << y1Dots << " m " << x2Dots << " " << y2Dots << " l s\n";
        Page << "Q\n";
    }

    int TXPBackendPortableDocumentFormat::PutBookmark(int PageNumber, const std::wstring& Caption, int ParentHandle)
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::PutBookmark\n";

        if(ParentHandle < 0 || ParentHandle >= (int)Bookmarks.size())
            ParentHandle = 0;
        PortableDocumentFormatBookmark* parent = Bookmarks[ParentHandle];
        Bookmarks.push_back(new PortableDocumentFormatBookmark(PageNumber, ToPdfSystemString(Caption), parent));
        return Bookmarks.size()-1;
    }

    void TXPBackendPortableDocumentFormat::FinishPage()
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::FinishPage\n";

        // Internal and External References
        if(PageLinks.size() > 0 || Hyperlinks.size() > 0)
        {
            (*stream) << "  /Annots [\n";
            for(list<PortableDocumentFormatPageLink>::iterator i = PageLinks.begin(); i != PageLinks.end(); i++)
            {
                (*stream) << "    << /Type /Annot /Subtype /Link "
                          << "/Dest [" << FirstPageObjectId + ObjectsPerPage*(i->page) << " 0 R /Fit] "
                          << "/Rect [" << i->x << " " << i->y << " " << (i->x + i->w) << " " << (i->y - i->h) << "] "
                          << "/BS << /S /S /Type /Border /W 1 >> /Border [0 0 0] >>\n";
            }
            for(list<PortableDocumentFormatHyperlink>::iterator i = Hyperlinks.begin(); i != Hyperlinks.end(); i++)
            {
                (*stream) << "    << /Type /Annot /Subtype /Link /A << /S /URI /URI (" << i->url << ") >> "
                          << "/Rect [" << i->x << " " << i->y << " " << (i->x+i->w) << " " << (i->y - i->h) << "] "
                          << "/BS << /S /S /Type /Border /W 1 >> /Border [0 0 0] >>\n";
            }

            (*stream) << "  ]\n";
            PageLinks.clear();
            Hyperlinks.clear();
        }



        (*stream) << "  /Resources <<"
                  << " /Font " << FontRegisterObjectId << " 0 R"
               // << "    /ProcSet [ /PDF /ImageC ]\n"
               // << "    /ExtGState << /GS1 " << ObjectCounter+1 << " 0 R >>\n"
                  << " >>\n";
        (*stream) << "  /Contents " << ObjectCounter+1 << " 0 R\n";
        (*stream) << ">>\n";
        (*stream) << "endobj \n\n";
        ObjectCounter++;



        // Object 4 is the Content of Page 1
        ObjectOffsets.push_back(stream->tellp());
        (*stream) << ObjectCounter << " 0 obj \n";
        (*stream) << "<< /Length " << ((int)(Page.tellp())) << " >>\n";
        (*stream) << "stream\n";
        (*stream) << Page.str();
        (*stream) << "\nendstream \n";
        (*stream) << "endobj \n\n% ===============================================================================\n\n";
        ObjectCounter++;


/*
 *      IF WE UNCOMMENT THIS (OR OTHERWISE CHANGE THE NUMBER OF OBJECTS
 *      PER PAGE, WE MUST ALSO CHANGE THE int ObjectsPerPage!!!!
 *
        // Object 5 is the "ExtGState" of Page 1 (whatever that is...)
        ObjectOffsets.push_back(stream->tellp());
        (*stream) << ObjectCounter << " 0 obj \n";
        (*stream) << "<<\n";
        (*stream) << "/Type /ExtGState\n";
        (*stream) << "/SA false\n";
        (*stream) << "/OP false\n";
        (*stream) << "/HT /default\n";
        (*stream) << ">>\n";
        (*stream) << "endobj \n";
        ObjectCounter++;
*/

        Page.str("");
        TXPBackendFile::FinishPage();
    }

    void TXPBackendPortableDocumentFormat::Finish()
    {
        if(DEBUG_EXPORT) cerr << "TXPBackendPortableDocumentFormat::Finish\n";

        // Object 5 is the Font-Register
        (*stream) << "% ==================================================================================== FONTS\n\n";
        ObjectOffsets.push_back(stream->tellp());
        (*stream) << ObjectCounter << " 0 obj \n<<\n";
        int fontid = 1;
        for(std::list<std::string>::iterator i = Fonts.begin(); i != Fonts.end(); i++)
        {
            (*stream) << "  /F" << fontid++ << " << /Type /Font /Subtype /Type1 /BaseFont /" << i->c_str();
            if(Codepage[*i].size() > 0)
            {
                (*stream) << " /Encoding << /Type /Encoding /Differences [ 1 \n   ";
                int left = 3;
                for(std::vector<wchar_t>::iterator j = Codepage[*i].begin(); j != Codepage[*i].end(); j++)
                {
                    std::string glyphname = TXPBackendPostscript::GlyphName(*j);
                    left += glyphname.size()+2;
                    if(left > 80)
                    {
                        (*stream) << "\n   ";
                        left = 5+glyphname.size();
                    }
                    (*stream) << " /" << glyphname;
                }
                (*stream) << "\n  ] >>";
            }
            (*stream) << " >> \n";
        }
        (*stream) << ">>\n";
        (*stream) << "endobj \n\n";
        ObjectCounter++;

        for(std::list<std::string>::iterator i = Fonts.begin(); i != Fonts.end(); i++)
        {
            ObjectOffsets.push_back(stream->tellp());
            ifstream fontfile(i->c_str(), ifstream::ate|ifstream::binary);
            (*stream) << ObjectCounter << " 0 obj \n<<\n"
                      << "/Type /EmbeddedFile /Length " << fontfile.tellg() << " >>\nstream\n";
            fontfile.seekg(ifstream::beg);
            //copy(istreambuf_iterator<wchar_t>(fontfile), istreambuf_iterator<wchar_t>(), ostreambuf_iterator<wchar_t>(*stream));
            fontfile.close();
            (*stream) << "\nendstream\nendobj\n\n";
            ObjectCounter++;
        }



         // Object 6 is the Outline-Root
        if(Bookmarks.size() > 1)
        {
            (*stream) << "% ================================================================================== OUTLINE\n\n";
            int bookmarkObjectCounter = 0;
            for(std::vector<PortableDocumentFormatBookmark*>::iterator i = Bookmarks.begin(); i != Bookmarks.end(); i++)
                (*i)->ObjectId = ObjectCounter + bookmarkObjectCounter++;
            for(std::vector<PortableDocumentFormatBookmark*>::iterator i = Bookmarks.begin(); i != Bookmarks.end(); i++)
            {
                ObjectOffsets.push_back(stream->tellp());
                (*stream) << ObjectCounter << " 0 obj \n<<";
                if((*i)->Parent != NULL)
                {
                    (*stream) << " /Parent " << (*i)->Parent->ObjectId << " 0 R"
                            << " /Title " << (*i)->Caption
                            << " /Dest [" << FirstPageObjectId + ObjectsPerPage*((*i)->Page) << " 0 R /Fit]";
                }
                if((*i)->FirstChild != NULL) // has children
                {
                    (*stream) << " /First " << (*i)->FirstChild->ObjectId << " 0 R"
                            << " /Last " << (*i)->LastChild->ObjectId << " 0 R";
                }
                if((*i)->NextSibling != NULL)
                    (*stream) << " /Next " << (*i)->NextSibling->ObjectId << " 0 R";
                if((*i)->PrevSibling != NULL)
                    (*stream) << " /Prev " << (*i)->PrevSibling->ObjectId << " 0 R";

                (*stream) << " >>\nendobj \n\n";
                ObjectCounter++;
            }
        }


        (*stream) << "% ===================================================================================== FOOT\n\n";
        // Object 7 is the Catalog
        ObjectOffsets.push_back(stream->tellp());
        CatalogObjectId = ObjectCounter;
        (*stream) << ObjectCounter << " 0 obj \n";
        (*stream) << "<<\n";
        (*stream) << "  /Type /Catalog\n";
        (*stream) << "  /Outlines " << Bookmarks[0]->ObjectId << " 0 R\n";
        (*stream) << "  /Pages " << PageTreeObjectId << " 0 R\n";
        (*stream) << ">>\n";
        (*stream) << "endobj \n\n";
        ObjectCounter++;



        // Cross-Reference Table
        int xref_offset = stream->tellp();
        (*stream) << "xref\n";
        (*stream) << "0 " << ObjectOffsets.size()+1 << "\n";
        (*stream) << "0000000000 65535 f \n";
        for(list<int>::iterator i = ObjectOffsets.begin(); i != ObjectOffsets.end(); i++)
            (*stream) << setfill('0') << setw(10) << (*i) << " 00000 n \n";



        // Trailer
        (*stream) << "trailer\n\n";
        (*stream) << "<<\n";
        if(InfoObjectId >= 0)
            (*stream) << "  /Info " << InfoObjectId << " 0 R\n";
        if(CatalogObjectId >= 0)
            (*stream) << "  /Root " << CatalogObjectId << " 0 R\n";
        (*stream) << "  /Size " << ObjectOffsets.size()+1 << "\n"; // number of entries in the xref table
        (*stream) << ">>\n";
        (*stream) << "startxref\n" << (xref_offset) << "\n";
        (*stream) << "%%EOF\n";

        TXPBackendFile::Finish();
    }
