%code requires {

    // Many thanks to
    // http://www.thebinaryidiot.com/archives/2010/03/06/flex-bison-cpp-example/

    #include "../../include/txp/txp.h"
    class TXPLexer;
    using namespace std;
}

%skeleton "lalr1.cc"
%defines
%define parser_class_name {TXPParser}
%parse-param { TXPLexer &scanner }
%parse-param { TXPElement** result }
%lex-param { TXPLexer &scanner }

%union
{
    std::wstring* stringval;
    float* floatval;
    int* intval;

    TXPElement* element;
    TXPNode* node;
    TXPStatement* statement;

    TXPArithmeticExpression* arithexpr;
    TXPBool* boolexpr;
    TXPRelation* relation;
};

%code {
    extern bool PARSER_DEBUG;
    static int yylex(yy::TXPParser::semantic_type* yylval, TXPLexer &scanner);
    void yy_pop_state(TXPLexer &scanner);
    void yy_push_state_code(TXPLexer &scanner);
}

%start S

%token END 0 "end of file"

%token OPEN
%token CLOSE
%token OPENBRACES
%token CLOSEBRACES
%token COMMA
%token SEMICOLON
%token IDENTIFIER

%token UNICODE
%token NEWPAGE
%token LABEL
%token PAGELINK
%token HYPERLINK
%token BOOKMARK

%token FOR
%token IF
%token ELSE
%token TRUE
%token FALSE

%token VOID

%token AND
%token OR
%token NOT

%token ASSIGN
%token EQ
%token NEQ
%token LE
%token LT
%token GT
%token GE

%token ADD
%token SUB
%token MUL
%token DIV
%token MOD
%token ADDASSIGN
%token SUBASSIGN
%token MULASSIGN
%token DIVASSIGN
%token MODASSIGN

%token INC
%token DEC
%token ADDA
%token SUBA
%token MULA
%token DIVA
%token MODA

%token <floatval> FLOAT
%token <intval> INT
%token <stringval> STRING
%token <stringval> CHAR

%left OR ADD SUB
%left AND MUL DIV MOD
%precedence NOT

%% // Produktionen

S:        ELEMENTS END                         { if(PARSER_DEBUG) cerr << "S: ELEMENTS END\n";
                                                 *result = $<node>1; }
       ;

ELEMENTS: ELEMENTS ELEMENT                     { if(PARSER_DEBUG) cerr << "ELEMENTS: ELEMENTS ELEMENT\n";
                                                 // the following code just unites consecutive "TXPString" Nodes
                                                 // (without this, we would have a TXPString instance for every char)
                                                 $<node>$ = NULL;
                                                 TXPNode *n = dynamic_cast<TXPNode*>($<node>1);
                                                 if(n != NULL)
                                                 {
                                                     TXPString *ns = dynamic_cast<TXPString*>(n->element);
                                                     TXPString *s = dynamic_cast<TXPString*>($<element>2);
                                                     if(s != NULL && ns != NULL)
                                                     {
                                                         *(ns->value) += *(s->value);
                                                         delete s;
                                                         $<node>$ = $<node>1;
                                                     }
                                                 }
                                                 if($<node>$ == NULL)
                                                     $<node>$ = new TXPNode($<node>1, $<element>2);
                                               }
        |                                      { if(PARSER_DEBUG) cerr << "ELEMENTS: empty\n";
                                                 $<node>$ = NULL; }
        ;

ELEMENT:  CHAR                                 { if(PARSER_DEBUG) cerr << "ELEMENT: CHAR\n"; $<element>$ = new TXPString($<stringval>1); }
        | CODEELEMENT                          { if(PARSER_DEBUG) cerr << "ELEMENT: CODEELEMENT\n"; $<element>$ = $<statement>1; }
        | OPENBRACES ELEMENTS CLOSEBRACES      { if(PARSER_DEBUG) cerr << "ELEMENT: OPENBRACES ELEMENTS CLOSEBRACES\n"; $<element>$ = new TXPBlock($<node>2); }
        | UNICODE OPEN INT CLOSE               { if(PARSER_DEBUG) cerr << "CODEELEMENT: UNICODE OPEN INT CLOSE\n";
                                                 yy_pop_state(scanner);
                                                 $<element>$ = new TXPString(new wstring(1, wchar_t(*$<intval>3)));
                                                 delete $<intval>3;
                                                 //$<statement>$ = new TXPUnicode($<intval>3);
                                                 }
        ;



SIGNEDFLOAT: FLOAT                             {  if(PARSER_DEBUG) cerr << "SIGNEDFLOAT: FLOAT\n";
                                                  $<floatval>$ = $<floatval>1; }
           | ADD FLOAT                         {  if(PARSER_DEBUG) cerr << "SIGNEDFLOAT: ADD FLOAT\n";
                                                  $<floatval>$ = $<floatval>2; }
           | SUB FLOAT                         {  if(PARSER_DEBUG) cerr << "SIGNEDFLOAT: SUB FLOAT\n";
                                                  $<floatval>$ = $<floatval>2; (*$<floatval>2) *= -1; }
           | SIGNEDINT                         {  if(PARSER_DEBUG) cerr << "SIGNEDFLOAT: SIGNEDINT\n";
                                                  $<floatval>$ = new float(*$<intval>1); delete $<intval>1;}
           ;
SIGNEDINT: INT                                 {  if(PARSER_DEBUG) cerr << "SIGNEDINT: INT\n";
                                                  $<intval>$ = $<intval>1; }
           | ADD INT                           {  if(PARSER_DEBUG) cerr << "SIGNEDINT: ADD INT\n";
                                                  $<intval>$ = $<intval>2; }
           | SUB INT                           {  if(PARSER_DEBUG) cerr << "SIGNEDINT: SUB INT\n";
                                                  $<intval>$ = $<intval>2; (*$<intval>2) *= -1;}
           ;



CODEELEMENT: FOR OPEN STATEMENT SEMICOLON BOOL SEMICOLON STATEMENT CLOSE { yy_pop_state(scanner); } ELEMENT
                                               { if(PARSER_DEBUG) cerr << "CODEELEMENT: FOR ...\n";
                                                 $<statement>$ = new TXPFor($<statement>3, $<boolexpr>5, $<statement>7, $<element>10); }

        | VOID IDENTIFIER OPEN CLOSE { yy_pop_state(scanner); } ELEMENT {
                                                 if(PARSER_DEBUG) cerr << "CODEELEMENT: VOID IDENTIFIER OPEN CLOSE ELEMENT\n";
                                                 $<statement>$ = new TXPProcedureDefinition($<stringval>2, $<element>6); }

        | IDENTIFIER OPEN CLOSE { yy_pop_state(scanner); } {
                                                 if(PARSER_DEBUG) cerr << "CODEELEMENT: IDENTIFIER OPEN CLOSE\n";
                                                 $<statement>$ = new TXPProcedureCall($<stringval>1); }

        | IFTHEN                               { yy_pop_state(scanner);
                                                 if(PARSER_DEBUG) cerr << "CODEELEMENT: IFTHEN\n";
                                                 $<statement>$ = $<statement>1;}

        | IFTHEN ELSE { yy_pop_state(scanner); } ELEMENT
                                               { if(PARSER_DEBUG) cerr << "CODEELEMENT: IFTHEN ELSE\n";
                                                 TXPIfThenElse* ifthen = dynamic_cast<TXPIfThenElse*>($<statement>1);
                                                 ifthen->elseblock = $<element>4;
                                                 $<statement>$ = ifthen;
                                               }
        | NEWPAGE OPEN SIGNEDFLOAT COMMA SIGNEDFLOAT CLOSE { if(PARSER_DEBUG) cerr << "CODEELEMENT: NEWPAGE OPEN SIGNEDFLOAT COMMA SIGNEDFLOAT CLOSE\n";
                                                 yy_pop_state(scanner);
                                                 $<statement>$ = new TXPNewpage($<floatval>3, $<floatval>5); }
        | LABEL OPEN IDENTIFIER CLOSE          { if(PARSER_DEBUG) cerr << "CODEELEMENT: LABEL OPEN IDENTIFIER CLOSE\n";
                                                 yy_pop_state(scanner);
                                                 $<statement>$ = new TXPLabel($<stringval>3); }
        | HYPERLINK OPEN SIGNEDFLOAT COMMA SIGNEDFLOAT COMMA SIGNEDFLOAT COMMA SIGNEDFLOAT COMMA IDENTIFIER CLOSE
                                               { if(PARSER_DEBUG) cerr << "CODEELEMENT: HYPERLINK OPEN SIGNEDFLOAT COMMA ... CLOSE\n";
                                                 yy_pop_state(scanner);
                                                 $<statement>$ = new TXPHyperlink($<floatval>3, $<floatval>5, $<floatval>7, $<floatval>9, $<stringval>11); }
        | BOOKMARK OPEN SIGNEDINT COMMA IDENTIFIER CLOSE
                                               { if(PARSER_DEBUG) cerr << "CODEELEMENT: BOOKMARK OPEN SIGNEDINT COMMA IDENTIFIER CLOSE\n";
                                                 yy_pop_state(scanner);
                                                 $<statement>$ = new TXPBookmark($<intval>3, $<stringval>5); }
        | PAGELINK OPEN SIGNEDFLOAT COMMA SIGNEDFLOAT COMMA SIGNEDFLOAT COMMA SIGNEDFLOAT COMMA IDENTIFIER CLOSE
                                               { if(PARSER_DEBUG) cerr << "CODEELEMENT: PAGELINK OPEN SIGNEDFLOAT COMMA ... CLOSE\n";
                                                 yy_pop_state(scanner);
                                                 $<statement>$ = new TXPPageLink($<floatval>3, $<floatval>5, $<floatval>7, $<floatval>9, $<stringval>11); }
        | STATEMENT SEMICOLON                  { if(PARSER_DEBUG) cerr << "CODEELEMENT: STATEMENT SEMICOLON\nyy_pop_state()\n";
                                                 $<statement>$ = $<statement>1;
                                                 yy_pop_state(scanner); }
        ;
IFTHEN: IF OPEN BOOL CLOSE { yy_pop_state(scanner);} ELEMENT
                                               { yy_push_state_code(scanner);
                                                 if(PARSER_DEBUG) cerr << "IFTHEN: IF ...\n";
                                                 $<statement>$ = new TXPIfThenElse($<boolexpr>3, $<element>6, NULL); }
      ;



STATEMENT: IDENTIFIER ASSIGN ARITH             { if(PARSER_DEBUG) cerr << "STATEMENT: IDENTIFIER ASSIGN ARITH\n";
                                                 $<statement>$ = new TXPAssignment($<stringval>1, $<arithexpr>3); }
        | IDENTIFIER ADDASSIGN ARITH           { if(PARSER_DEBUG) cerr << "STATEMENT: IDENTIFIER ADD ASSIGN ARITH\n";
                                                 $<statement>$ = new TXPAssignment($<stringval>1, new TXPArithmeticPlus(new TXPArithmeticVariable(new std::wstring(*$<stringval>1)), $<arithexpr>3)); }
        | IDENTIFIER SUBASSIGN ARITH           { if(PARSER_DEBUG) cerr << "STATEMENT: IDENTIFIER SUB ASSIGN ARITH\n";
                                                 $<statement>$ = new TXPAssignment($<stringval>1, new TXPArithmeticMinus(new TXPArithmeticVariable(new std::wstring(*$<stringval>1)), $<arithexpr>3)); }
        | IDENTIFIER MULASSIGN ARITH           { if(PARSER_DEBUG) cerr << "STATEMENT: IDENTIFIER MUL ASSIGN ARITH\n";
                                                 $<statement>$ = new TXPAssignment($<stringval>1, new TXPArithmeticMult(new TXPArithmeticVariable(new std::wstring(*$<stringval>1)), $<arithexpr>3)); }
        | IDENTIFIER DIVASSIGN ARITH           { if(PARSER_DEBUG) cerr << "STATEMENT: IDENTIFIER DIV ASSIGN ARITH\n";
                                                 $<statement>$ = new TXPAssignment($<stringval>1, new TXPArithmeticDiv(new TXPArithmeticVariable(new std::wstring(*$<stringval>1)), $<arithexpr>3)); }
        | IDENTIFIER MODASSIGN ARITH           { if(PARSER_DEBUG) cerr << "STATEMENT: IDENTIFIER MOD ASSIGN ARITH\n";
                                                 $<statement>$ = new TXPAssignment($<stringval>1, new TXPArithmeticMod(new TXPArithmeticVariable(new std::wstring(*$<stringval>1)), $<arithexpr>3)); }
        | ARITH                                { if(PARSER_DEBUG) cerr << "STATEMENT: ARITH\n";
                                                 $<statement>$ = $<arithexpr>1; }
        ;




BOOL:     TRUE                                 { if(PARSER_DEBUG) cerr << "BOOL: TRUE\n";            $<boolexpr>$ = new TXPBoolTrue(); }
        | FALSE                                { if(PARSER_DEBUG) cerr << "BOOL: FALSE\n";           $<boolexpr>$ = new TXPBoolFalse(); }
        | BOOL AND BOOL                        { if(PARSER_DEBUG) cerr << "BOOL: BOOL AND BOOL\n";   $<boolexpr>$ = new TXPBoolAnd($<boolexpr>1, $<boolexpr>3); }
        | BOOL OR BOOL                         { if(PARSER_DEBUG) cerr << "BOOL: BOOL OR BOOL\n";    $<boolexpr>$ = new TXPBoolOr($<boolexpr>1, $<boolexpr>3); }
        | NOT BOOL                             { if(PARSER_DEBUG) cerr << "BOOL: NOT BOOL\n";        $<boolexpr>$ = new TXPBoolNot($<boolexpr>2); }
        | OPEN BOOL CLOSE                      { if(PARSER_DEBUG) cerr << "BOOL: OPEN BOOL CLOSE\n"; $<boolexpr>$ = $<boolexpr>2; }
        | RELATION                             { if(PARSER_DEBUG) cerr << "BOOL: RELATION\n";        $<boolexpr>$ = $<relation>1; }
        ;

RELATION: ARITH LE ARITH                       { if(PARSER_DEBUG) cerr << "RELATION: ARITH LE ARITH\n";  $<relation>$ = new TXPRelationLE($<arithexpr>1, $<arithexpr>3); }
        | ARITH LT ARITH                       { if(PARSER_DEBUG) cerr << "RELATION: ARITH LT ARITH\n";  $<relation>$ = new TXPRelationLT($<arithexpr>1, $<arithexpr>3); }
        | ARITH GT ARITH                       { if(PARSER_DEBUG) cerr << "RELATION: ARITH GT ARITH\n";  $<relation>$ = new TXPRelationGT($<arithexpr>1, $<arithexpr>3); }
        | ARITH GE ARITH                       { if(PARSER_DEBUG) cerr << "RELATION: ARITH GE ARITH\n";  $<relation>$ = new TXPRelationGE($<arithexpr>1, $<arithexpr>3); }
        | ARITH NEQ ARITH                      { if(PARSER_DEBUG) cerr << "RELATION: ARITH NEQ ARITH\n"; $<relation>$ = new TXPRelationNEQ($<arithexpr>1, $<arithexpr>3); }
        | ARITH EQ ARITH                       { if(PARSER_DEBUG) cerr << "RELATION: ARITH EQ ARITH\n";  $<relation>$ = new TXPRelationEQ($<arithexpr>1, $<arithexpr>3); }
        ;

ARITH:    ARITH ADD ARITH                      { if(PARSER_DEBUG) cerr << "ARITH: ARITH ADD ARITH\n";  $<arithexpr>$ = new TXPArithmeticPlus($<arithexpr>1, $<arithexpr>3); }
        | ARITH SUB ARITH                      { if(PARSER_DEBUG) cerr << "ARITH: ARITH SUB ARITH\n";  $<arithexpr>$ = new TXPArithmeticMinus($<arithexpr>1, $<arithexpr>3); }
        | ARITH MUL ARITH                      { if(PARSER_DEBUG) cerr << "ARITH: ARITH MUL ARITH\n";  $<arithexpr>$ = new TXPArithmeticMult($<arithexpr>1, $<arithexpr>3); }
        | ARITH DIV ARITH                      { if(PARSER_DEBUG) cerr << "ARITH: ARITH DIV ARITH\n";  $<arithexpr>$ = new TXPArithmeticDiv($<arithexpr>1, $<arithexpr>3); }
        | ARITH MOD ARITH                      { if(PARSER_DEBUG) cerr << "ARITH: ARITH MOD ARITH\n";  $<arithexpr>$ = new TXPArithmeticMod($<arithexpr>1, $<arithexpr>3); }
        | OPEN ARITH CLOSE                     { if(PARSER_DEBUG) cerr << "ARITH: OPEN ARITH CLOSE\n"; $<arithexpr>$ = $<arithexpr>2; }
        | SIGNEDFLOAT                          { if(PARSER_DEBUG) cerr << "ARITH: SIGNEDFLOAT\n";      $<arithexpr>$ = new TXPArithmeticConstant( $<floatval>1 ); }
        | IDENTIFIER                           { if(PARSER_DEBUG) cerr << "ARITH: IDENTIFIER\n";       $<arithexpr>$ = new TXPArithmeticVariable( $<stringval>1 ); }
        ;



%%

#include "../../include/parser/txp_lexer.h"

void yy::TXPParser::error(const std::string &msg)
{
    std::cerr << "Error near line " << scanner.GetCurrentLineNumber() << ": " << msg << std::endl;
}

static int yylex(yy::TXPParser::semantic_type* yylval, TXPLexer &scanner)
{
    return scanner.yylex(yylval);
}

void yy_pop_state(TXPLexer& scanner)
{
    scanner.pop_state();
}

void yy_push_state_code(TXPLexer& scanner)
{
    scanner.push_state_code();
}
