
    #include "../../include/txp/txp.h"

    using namespace std;
    extern bool DEBUG_EXECUTION;






    TXPContext::TXPContext(TXPMiddleEnd* middleend)
    {
        this->middleend = middleend;
    }

    TXPContext::~TXPContext()
    {
    }





    TXPElement::TXPElement()
    {
    }

    TXPElement::~TXPElement()
    {
    }




    TXPNode::TXPNode(TXPNode* prev, TXPElement* element)
        : TXPElement()
    {
        this->prev = prev;
        this->element = element;
    }

    TXPNode::~TXPNode()
    {
        delete prev;
        delete element;
    }

    void TXPNode::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPNode::Execute\n";
        if(prev != NULL)
            prev->Execute(context);
        if(element != NULL)
            element->Execute(context);
    }





    TXPBlock::TXPBlock(TXPNode* block)
    {
        this->block = block;
    }
    TXPBlock::~TXPBlock()
    {
        delete block;
    }

    void TXPBlock::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPBlock::Execute\n";
        if(block != NULL)
            block->Execute(context);
    }









    TXPStatement::TXPStatement()
        : TXPElement()
    {
    }
    TXPStatement::~TXPStatement()
    {
    }

    // =========================================================================================================
    // =========================================================================================================
    // =========================================================================================================


    TXPBool::TXPBool()
    {
    }
    TXPBool::~TXPBool()
    {
    }


    TXPBoolTrue::TXPBoolTrue()
        : TXPBool()
    {
    }
    TXPBoolTrue::~TXPBoolTrue()
    {
    }
    bool TXPBoolTrue::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPBoolTrue::Evaluate\n";

        return true;
    }


    TXPBoolFalse::TXPBoolFalse()
        : TXPBool()
    {
    }
    TXPBoolFalse::~TXPBoolFalse()
    {
    }
    bool TXPBoolFalse::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPBoolFalse::Evaluate\n";
        return false;
    }


    TXPBoolAnd::TXPBoolAnd(TXPBool* a, TXPBool* b)
        : TXPBool()
    {
        this->a = a;
        this->b = b;
    }
    TXPBoolAnd::~TXPBoolAnd()
    {
        delete a;
        delete b;
    }
    bool TXPBoolAnd::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPBoolAnd::Evaluate\n";

        return a->Evaluate(context) && b->Evaluate(context);
    }



    TXPBoolOr::TXPBoolOr(TXPBool* a, TXPBool* b)
        : TXPBoolAnd(a,b)
    {
    }
    TXPBoolOr::~TXPBoolOr()
    {
    }
    bool TXPBoolOr::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPBoolOr::Evaluate\n";

        return a->Evaluate(context) || b->Evaluate(context);
    }



    TXPBoolNot::TXPBoolNot(TXPBool* a)
        : TXPBool()
    {
        this->a = a;
    }
    TXPBoolNot::~TXPBoolNot()
    {
        delete a;
    }
    bool TXPBoolNot::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPBoolNot::Evaluate\n";

        return ! a->Evaluate(context);
    }


    // =========================================================================================================
    // =========================================================================================================
    // =========================================================================================================


    TXPRelation::TXPRelation(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
    {
        this->a = a;
        this->b = b;
    }

    TXPRelation::~TXPRelation()
    {
        delete a;
        delete b;
    }

    bool TXPRelation::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPRelation::Evaluate\n";
        return Combine(a->Evaluate(context), b->Evaluate(context));
    }



    TXPRelationEQ::TXPRelationEQ(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPRelation(a,b)
    {
    }
    TXPRelationEQ::~TXPRelationEQ()
    {
    }
    bool TXPRelationEQ::Combine(float a, float b)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPRelationEQ::Combine\n";
        return a == b;
    }


    TXPRelationNEQ::TXPRelationNEQ(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPRelation(a,b)
    {
    }
    TXPRelationNEQ::~TXPRelationNEQ()
    {
    }
    bool TXPRelationNEQ::Combine(float a, float b)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPRelationNEQ::Combine\n";
        return a != b;
    }


    TXPRelationLT::TXPRelationLT(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPRelation(a,b)
    {
    }
    TXPRelationLT::~TXPRelationLT()
    {
    }
    bool TXPRelationLT::Combine(float a, float b)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPRelationLT::Combine\n";
        return a < b;
    }


    TXPRelationLE::TXPRelationLE(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPRelation(a,b)
    {
    }
    TXPRelationLE::~TXPRelationLE()
    {
    }
    bool TXPRelationLE::Combine(float a, float b)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPRelationLE::Combine\n";
        return a <= b;
    }


    TXPRelationGE::TXPRelationGE(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPRelation(a,b)
    {
    }
    TXPRelationGE::~TXPRelationGE()
    {
    }
    bool TXPRelationGE::Combine(float a, float b)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPRelationGE::Combine\n";
        return a >= b;
    }


    TXPRelationGT::TXPRelationGT(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPRelation(a,b)
    {
    }
    TXPRelationGT::~TXPRelationGT()
    {
    }
    bool TXPRelationGT::Combine(float a, float b)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPRelationGT::Combine\n";
        return a > b;
    }


    // =========================================================================================================
    // =========================================================================================================
    // =========================================================================================================



    TXPArithmeticExpression::TXPArithmeticExpression()
        : TXPStatement()
    {
    }
    TXPArithmeticExpression::~TXPArithmeticExpression()
    {
    }
    void TXPArithmeticExpression::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticExpression::Execute\n";

        std::wstringstream ws;
        ws << Evaluate(context);
        context->middleend->AddText( ws.str() );
    }


    TXPArithmeticPlus::TXPArithmeticPlus(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
    {
        this->a = a;
        this->b = b;
    }
    TXPArithmeticPlus::~TXPArithmeticPlus()
    {
        delete a;
        delete b;
    }
    float TXPArithmeticPlus::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticPlus::Evaluate\n";
        return a->Evaluate(context) + b->Evaluate(context);
    }


    TXPArithmeticMinus::TXPArithmeticMinus(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPArithmeticPlus(a,b)
    {
    }
    TXPArithmeticMinus::~TXPArithmeticMinus()
    {
    }
    float TXPArithmeticMinus::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticMinus::Evaluate\n";
        return a->Evaluate(context) - b->Evaluate(context);
    }


    TXPArithmeticMult::TXPArithmeticMult(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPArithmeticPlus(a,b)
    {
    }
    TXPArithmeticMult::~TXPArithmeticMult()
    {
    }
    float TXPArithmeticMult::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticMult::Evaluate\n";
        return a->Evaluate(context) * b->Evaluate(context);
    }


    TXPArithmeticDiv::TXPArithmeticDiv(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPArithmeticPlus(a,b)
    {
    }
    TXPArithmeticDiv::~TXPArithmeticDiv()
    {
    }
    float TXPArithmeticDiv::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticDiv::Evaluate\n";
        return a->Evaluate(context) / b->Evaluate(context);
    }


    TXPArithmeticMod::TXPArithmeticMod(TXPArithmeticExpression* a, TXPArithmeticExpression* b)
        : TXPArithmeticPlus(a,b)
    {
    }
    TXPArithmeticMod::~TXPArithmeticMod()
    {
    }
    float TXPArithmeticMod::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticMod::Evaluate\n";
        float A = a->Evaluate(context);
        float B = b->Evaluate(context);
        return A - floor(A/B)*B;
    }


    TXPArithmeticVariable::TXPArithmeticVariable(std::wstring* name)
        : TXPArithmeticExpression()
    {
        this->name = name;
    }
    TXPArithmeticVariable::~TXPArithmeticVariable()
    {
        delete name;
    }
    float TXPArithmeticVariable::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticVariable::Evaluate\n";
        return context->FloatVars[*name];
    }


    TXPArithmeticConstant::TXPArithmeticConstant(float* value)
        : TXPArithmeticExpression()
    {
        this->value = value;
    }
    TXPArithmeticConstant::~TXPArithmeticConstant()
    {
        delete value;
    }
    float TXPArithmeticConstant::Evaluate(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPArithmeticConstant::Evaluate\n";
        return *value;
    }


    

    // =========================================================================================================
    // =========================================================================================================
    // =========================================================================================================




    TXPProcedureDefinition::TXPProcedureDefinition(std::wstring* name, TXPElement* block)
        : TXPStatement()
    {
        this->name = name;
        this->block = block;
    }

    TXPProcedureDefinition::~TXPProcedureDefinition ()
    {
        delete name;
        delete block;
    }

    void TXPProcedureDefinition::Execute(TXPContext* context)
    {
        if(context->Procedures.find(*name) == context->Procedures.end())
            context->Procedures[*name] = this;
        else
            std::cerr << "Procedure already defined.";
    }




    TXPProcedureCall::TXPProcedureCall(std::wstring* name)
        : TXPStatement()
    {
        this->name = name;
    }

    TXPProcedureCall::~TXPProcedureCall()
    {
        delete name;
    }

    void TXPProcedureCall::Execute(TXPContext* context)
    {
        if(context->Procedures.find(*name) != context->Procedures.end())
            context->Procedures[*name]->block->Execute(context);
        else
            std::cerr << "Call to undefined Procedure.";
    }




    TXPAssignment::TXPAssignment(std::wstring* name, TXPArithmeticExpression* expr)
        : TXPStatement()
    {
        this->name = name;
        this->expr = expr;
    }
    TXPAssignment::~TXPAssignment()
    {
        delete name;
        delete expr;
    }
    void TXPAssignment::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPAssignment::Execute\n";
        context->FloatVars[*name] = expr->Evaluate(context);
    }




    TXPFor::TXPFor(TXPStatement* a, TXPBool* cond, TXPStatement* b, TXPElement* block)
        : TXPStatement()
    {
        this->a = a;
        this->cond = cond;
        this->b = b;
        this->block = block;
    }
    TXPFor::~TXPFor()
    {
        delete a;
        delete cond;
        delete b;
        delete block;
    }
    void TXPFor::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPFor::Execute\n";

        if(a != NULL)
            a->Execute(context);
        while(cond == NULL || cond->Evaluate(context))
        {
            if(block != NULL)
                block->Execute(context);
            if(b != NULL)
                b->Execute(context);
        }
    }




    TXPIfThenElse::TXPIfThenElse(TXPBool* cond, TXPElement* thenblock, TXPElement* elseblock)
        : TXPStatement()
    {
        this->cond = cond;
        this->thenblock = thenblock;
        this->elseblock = elseblock;
    }
    TXPIfThenElse::~TXPIfThenElse()
    {
        delete cond;
        delete thenblock;
        delete elseblock;
    }
    void TXPIfThenElse::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPIfThenElse::Execute\n";

        if(cond->Evaluate(context))
        {
            if(thenblock != NULL)
                thenblock->Execute(context);
        }
        else
        {
            if(elseblock != NULL)
                elseblock->Execute(context);
        }
    }


    TXPUnicode::TXPUnicode(int* codepoint)
        : TXPStatement()
    {
        this->codepoint = codepoint;
    }
    TXPUnicode::~TXPUnicode()
    {
        delete codepoint;
    }
    void TXPUnicode::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPUnicode::Execute\n";

        context->middleend->AddText( wstring(1, wchar_t(*codepoint)) );
    }


    TXPNewpage::TXPNewpage(float* width, float* height)
        : TXPStatement()
    {
        this->width = width;
        this->height = height;
    }
    TXPNewpage::~TXPNewpage()
    {
        delete width;
        delete height;
    }
    void TXPNewpage::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPNewpage::Execute\n";

        context->middleend->NewPage(*width, *height);
    }




    TXPLabel::TXPLabel(std::wstring* name)
        : TXPStatement()
    {
        this->name = name;
    }
    TXPLabel::~TXPLabel()
    {
        delete name;
    }
    void TXPLabel::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPLabel::Execute\n";

        context->middleend->AddLabel(*name);
    }




    TXPPageLink::TXPPageLink(float* left, float* top, float* width, float* height, std::wstring* name)
        : TXPStatement()
    {
        this->left = left;
        this->top = top;
        this->width = width;
        this->height = height;
        this->name = name;
    }
    TXPPageLink::~TXPPageLink()
    {
        delete left;
        delete top;
        delete width;
        delete height;
        delete name;
    }
    void TXPPageLink::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPPageLink::Execute\n";

        context->middleend->AddPageLink(*left, *top, *width, *height, *name);
    }



    TXPHyperlink::TXPHyperlink(float* left, float* top, float* width, float* height, std::wstring* name)
        : TXPPageLink(left, top, width, height, name)
    {
    }
    TXPHyperlink::~TXPHyperlink()
    {
    }
    void TXPHyperlink::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPHyperlink::Execute\n";

        context->middleend->AddHyperlink(*left, *top, *width, *height, *name);
    }



    TXPBookmark::TXPBookmark(int* depthRelativeToPrevious, std::wstring* Caption)
        : TXPStatement()
    {
        this->depthRelativeToPrevious = depthRelativeToPrevious;
        this->Caption = Caption;
    }
    TXPBookmark::~TXPBookmark()
    {
        delete depthRelativeToPrevious;
        delete Caption;
    }
    void TXPBookmark::Execute(TXPContext* context)
    {
        if(DEBUG_EXECUTION)
            cerr << "TXPBookmark::Execute\n";

        context->middleend->AddBookmark(*depthRelativeToPrevious, *Caption);
    }







